import os
import socket
import time
import contextlib
import logging

import qtrace

from . import Analyzer

l = logging.getLogger("archr.analyzers.qtrace")


class QTraceAnalyzer(Analyzer):
    REQUIRED_IMPLANT = "qtrace"

    def fire(self, machine_type=None, **kwargs):
        if machine_type is None:
            machine_type = qtrace.TraceMachine

        fire_path = os.path.join(self.target.tmpwd, "qtrace", "fire")
        args_prefix = [fire_path, "--"]

        with self.target.flight_context(args_prefix=args_prefix, **kwargs) as flight:
            process = flight.process
            argv = self.target.target_args

            def start(machine):
                host = self.target.ipv4_address
                machine.trace_socket = qtrace.create_connection((host, 4242), sleep_time=0.1)
                machine.gdb = machine.gdb_client((host, 1234), machine)
                machine.std_streams = (
                    process.stdin,
                    process.stdout,
                    process.stderr,
                )

            machine = machine_type(argv)

            machine.start = start.__get__(machine)

            machine.run()

            return machine



import tempfile
import logging
import angr
import cle
import os

from angr.simos import SimUserland
from elftools.elf.elffile import ELFFile
from elftools.elf.segments import NoteSegment

l = logging.getLogger("archr.analyzers.angr")

from . import Analyzer


def _extract_mapping_from_core_file(core_path):
    with open(core_path, "rb") as core_file:
        core_elf = ELFFile(core_file)
        for segment in core_elf.iter_segments():
            if isinstance(segment, NoteSegment):
                notes = list(segment.iter_notes())
                break
        else:
            l.warning(
                "The core file @ %s does not contain a NOTE segment, you should fix that. We "
                "will attempt to get the mapped addresses from the scout_analyzer instead I guess.",
                core_path,
            )
            return None
    for note in notes:
        if note["n_type"] == "NT_FILE":
            file_note = note
            break
    else:
        l.warning(
            "Could not find an NT_FILE note in the core file @ %s. This is bad, and you should fix "
            "it. We will fall back to trying to extract the mappings from the scout_analyzer.",
            core_path,
        )
        return None

    mappings = file_note["n_desc"]["Elf_Nt_File_Entry"]
    paths = [p.decode() for p in file_note["n_desc"]["filename"]]
    assert len(paths) == len(mappings)
    mem_map = {}
    for fpath, mapping in zip(paths, mappings):
        mem_map[fpath] = min(mem_map.get(fpath, mapping["vm_start"]), mapping["vm_start"])

    return mem_map


class angrProjectAnalyzer(Analyzer):
    """
    Constructs an angr project to match the target precisely
    """

    def __init__(self, target, scout_analyzer=None, custom_hooks=None, custom_systemcalls=None, static_simproc=False):
        # pylint:disable=too-many-arguments
        """

        :param target:          The target to work on.
        :param scout_analyzer:      The scout analyzer.
        :param static_simproc:  When enabled, angr will hook functions in the main binary with SimProcedures if
                                available. This is useful when dealing with statically linked binaries.
        :type static_simproc:   bool
        """

        super().__init__(target)
        self.scout_analyzer = scout_analyzer
        self.static_simproc = static_simproc

        self.custom_hooks = {} if custom_hooks is None else dict(custom_hooks)
        self.custom_syscalls = {} if custom_systemcalls is None else dict(custom_systemcalls)

        self.project = None
        self._mem_mapping = None
        self._lib_folder = None

    def fire(
        self, core_path=None, return_loader=False, project_kwargs=None, **cle_args
    ):  # pylint:disable=arguments-differ
        # if the the project is already created, return what the user wants
        if self.project is not None:
            return self.project if not return_loader else self.project.loader

        auto_load_libs = cle_args.get("auto_load_libs", True)

        # from now on, try to create a angr project
        if project_kwargs is None:
            project_kwargs = dict(auto_load_libs=auto_load_libs)

        # TODO: this introduce file leak. However, we probably need some redesign to fix it
        tmpdir = tempfile.mkdtemp(prefix="archr_angr_project_analyzer")
        self._lib_folder = tmpdir
        self.target.retrieve_into(self.target.target_path, tmpdir)
        the_binary = os.path.join(tmpdir, os.path.basename(self.target.target_path))

        # preload the binary to decide if it supports setting library options or base addresses
        cle_args.update(cle_args.pop("load_options", {}))
        cle_args.pop("use_sim_procedures", None)  # TODO do something less hacky than this
        preload_kwargs = dict(cle_args)
        preload_kwargs["auto_load_libs"] = False
        preloader = cle.Loader(the_binary, **preload_kwargs)

        self._mem_mapping = None

        if core_path is not None:
            self._mem_mapping = _extract_mapping_from_core_file(core_path)

        if self._mem_mapping is None and self.scout_analyzer is not None:
            _, _, _, self._mem_mapping = self.scout_analyzer.fire()

        self._mem_mapping = self._mem_mapping or {}

        target_libs = [lib for lib in self._mem_mapping if lib.startswith("/")]
        the_libs = []
        for target_lib in target_libs:
            local_lib = os.path.join(tmpdir, os.path.basename(target_lib))
            self.target.retrieve_into(target_lib, tmpdir)
            if auto_load_libs:
                the_libs.append(local_lib)
        lib_opts = {os.path.basename(lib): {"base_addr": libaddr} for lib, libaddr in self._mem_mapping.items()}
        bin_opts = {}
        if preloader.main_object.pic:
            bin_opts = lib_opts.get(os.path.basename(self.target.target_path), {})

        # if a core dump is specified, create a project based on the core dump
        if core_path:
            file_mapping = {}

            # grab remote libraries to local machine and build the mapping
            for remote_path in self._mem_mapping:
                # use heuristic to distinguish file mappings from others
                if not remote_path.startswith("/"):
                    continue
                self.target.retrieve_into(remote_path, tmpdir)
                local_path = os.path.join(tmpdir, os.path.basename(remote_path))
                file_mapping[remote_path] = local_path

            bin_opts = {"backend": "elfcore", "executable": the_binary, "remote_file_mapping": file_mapping}
            self.project = angr.Project(core_path, main_opts=bin_opts, rebase_granularity=0x1000, **project_kwargs)
            if not return_loader:
                self._apply_all_hooks()
            self.project.loader._main_object = self.project.loader.elfcore_object._main_object
            return self.project if not return_loader else self.project.loader

        if return_loader:
            return cle.Loader(the_binary, preload_libs=the_libs, lib_opts=lib_opts, main_opts=bin_opts, **cle_args)
        self.project = angr.Project(
            the_binary, preload_libs=the_libs, lib_opts=lib_opts, main_opts=bin_opts, **project_kwargs
        )

        self._apply_all_hooks()

        if return_loader:
            return self.project.loader
        return self.project

    def _apply_all_hooks(self):
        if self.static_simproc:
            self._apply_simprocedures()

        for location, hook in self.custom_hooks.items():
            if type(location) is str:
                self.project.hook_symbol(location, hook)
                l.debug("Hooking symbol %s -> %s...", location, hook.display_name)
            else:
                self.project.hook(location, hook)
                l.debug("Hooking %s -> %s...", location, hook.display_name)

        if self.custom_syscalls:
            assert isinstance(self.project.simos, SimUserland)

            for name, sys_sim in self.custom_syscalls.items():
                l.debug("Hooking system call %s with %s", name, sys_sim)
                self.project.simos.syscall_library.procedures[name] = sys_sim

    def _apply_simprocedures(self):
        """
        Apply SimProcedures to functions inside the main binary.

        :return: None
        """

        # all SimProcedures indexed by name, assuming no name conflicts exist
        simprocs = {}
        for _, simproc_dict in angr.SIM_PROCEDURES.items():
            for name, simproc in simproc_dict.items():
                simprocs[name] = simproc

        for sym in self.project.loader.main_object.symbols:
            if sym.type == cle.SymbolType.TYPE_FUNCTION and sym.name in simprocs:
                l.debug("Hooking symbol %s in binary %s.", sym.name, self.project.filename)
                self.project.hook_symbol(sym.name, simprocs[sym.name]())



import subprocess
import struct
import re
import logging
import cle
import io

from . import strace_parser

l = logging.getLogger("archr.utils")


def parse_ldd(mem_map_str):
    entries = [l.strip() for l in mem_map_str.decode("utf-8").splitlines()]
    parsed = {}
    for entry in entries:
        if "=>" in entry:
            libname, paren_addr = entry.split("=>")[1].split()
        else:
            libname, paren_addr = entry.split()
        libaddr = int(paren_addr.strip("()"), 16)
        parsed[libname] = libaddr
    return parsed


def parse_proc_maps(proc_str):
    entries = [l.strip() for l in proc_str.splitlines()]
    parsed = {}
    for entry in entries:
        what = entry.split()[-1].decode("utf-8")
        addr_range = entry.split()[0]
        start, end = addr_range.split(b"-")
        if what in parsed:
            continue
        if what.startswith("/"):
            parsed[what] = int(start, 16)
        elif what.startswith("["):
            parsed[what] = int(start, 16)
            parsed[what.rstrip("]") + "-end]"] = int(end, 16)
    return parsed


def lib_dependencies(filepath):
    mem_map_str, _ = subprocess.Popen(["ldd", filepath], stdout=subprocess.PIPE).communicate()
    return [lib for lib in parse_ldd(mem_map_str) if lib != "linux-vdso.so.1"]


def hook_entry(binary, asm_code=None, bin_code=None):
    main_bin = io.BytesIO(binary)
    b = cle.Loader(main_bin, auto_load_libs=False, perform_relocations=False)
    start_addr = b.main_object.addr_to_offset(b.main_object.entry)
    arch = b.main_object.arch
    if arch.name in ("ARMHF", "ARMEL") and arch.is_thumb(start_addr):  # OMG, thumb mode is a disaster
        start_addr &= ~1  # recover the real address
        main_bin.seek(start_addr)
        padding = (4 - (start_addr + 8) % 4) % 4  # we hardcode the shellcode so that its length is 8

        # we can' use arch.asm here because the shellcode THUMB, 8+padding-4 because the shellcode has length 8+padding,
        # we also need to take into account that in arm, pc points to two instructions ahead, which is 4 bytes in thumb mode
        main_bin.write(b"xF\x00\xf1" + struct.pack("<H", 8 + padding - 4) + b"\x00G" + b"A" * padding)

        # now place our payload after this mini shellcode
        start_addr += 8 + padding
    main_bin.seek(start_addr)
    main_bin.write(b.main_object.arch.asm(asm_code) if asm_code else bin_code)
    main_bin.seek(0)
    return main_bin.read()


def filter_strace_output(lines):
    """
    a function to filter QEMU logs returning only the strace entries

    Parameters
    ----------
    lines : list
        a list of strings representing the lines from a QEMU log/trace.

    Returns
    -------
    list
        a list of strings representing only the strace log entries
        the entries will also be cleaned up if a page dump occurs in the middle of them
    """

    # we only want the strace lines, so remove/ignore lines that start with the following:
    line_starts = [
        r"^[\d,a-f]{16}-",  # pylint: disable=anomalous-backslash-in-string
        r"^[\d,a-f]{8}-",  # pylint: disable=anomalous-backslash-in-string
        "^page",
        "^start",
        "^host",
        "^Locating",
        "^guest_base",
        "^end_",
        "^brk",
        "^entry",
        "^argv_",
        "^env_",
        "^auxv_",
        "^Trace",
        "^--- SIGSEGV",
        "^qemu",
    ]
    filter_string = "|".join(line_starts)

    filtered = []
    prev_line = ""
    for line in lines:
        if re.match(filter_string, line):
            continue
        # workaround for https://gitlab.com/qemu-project/qemu/-/issues/654
        if re.search("page layout changed following target_mmap", line):
            prev_line = line.replace("page layout changed following target_mmap", "")
            continue
        # The strace_parser does not properly handle _newselect strace lines (It's format differs significantly).
        # so, workaround by simply removing those lines. We currently do not use them.
        if re.search("newselect", line):
            continue  # ignore _newselect syscalls
        if re.match("^ = |^= ", line):
            line = prev_line + line

        filtered.append(line)
    return filtered


def get_file_maps(strace_log_lines):
    """
    a function to return a mapping of filenames and associated mmaped addressed for process under QEMU
    this is accomplished by tracking the filenames through file-desciptors across mmap() calls.

    Parameters
    ----------
    strace_log_lines : list
        a list of strings representing only the strace info logged by QEMU

    Returns
    -------
    dict
        a dictionary of filenames and mmapped addresses associated with each file

    """
    files = {"open": {}, "closed": {}, "sockets": []}

    entries = strace_parser.parse(strace_log_lines)
    entries = [entry for entry in entries if entry.syscall in ("openat", "open", "socket", "mmap", "mmap2", "close")]

    for entry in entries:
        # for an openat, create a dict entry for the file descriptor
        # the entry should be a tuple of the filename, and mmaps (initially empty)
        if entry.syscall == "openat":
            fd = entry.syscall.result
            # only care about file descriptors other than STDIN,STDOUT,STDERR
            # also ignore errors
            if fd >= 3:
                # use only the base filename
                filename = entry.syscall.args[1].split("/")[-1]
                # tracking if an executable page was ever mapped from the file descriptor
                files["open"][fd] = [filename, []]

        elif entry.syscall == "open":
            fd = entry.syscall.result
            # only care about file descriptors other than STDIN,STDOUT,STDERR
            # also ignore errors
            if fd >= 3:
                # use only the base filename
                filename = entry.syscall.args[0].split("/")[-1]
                # tracking if an executable page was ever mapped from the file descriptor
                files["open"][fd] = [filename, []]

        elif entry.syscall == "socket":
            fd = entry.syscall.result
            if fd >= 3:
                files["sockets"].append(fd)

        # if a file descriptor is closed, we need to remove it from the open files dictionary
        # we want to track the mmaps, so move it to 'closed' by file name since the file descriptor can be re-used.
        elif entry.syscall == "close":
            fd = entry.syscall.args[0]
            # only care about file descriptors other than STDIN,STDOUT,STDERR
            if fd >= 3:
                if fd in files["sockets"]:
                    files["sockets"].remove(fd)
                else:
                    filename = files["open"][fd][0]
                    mmaps = files["open"][fd][1]

                    # if we never mapped any pages, then we don't care about it.
                    if mmaps:
                        # otherwise move to 'closed'
                        files["closed"][filename] = mmaps

                    del files["open"][fd]

        # we can use the file descriptor to look up the dict entry to update the mmaps
        # TODO: track sizes which should be capturable from the mmap arguments
        elif entry.syscall in ("mmap", "mmap2"):
            # only care about valid file descriptors
            fd = entry.syscall.args[4]
            if fd >= 3:
                files["open"][fd][1].append(entry.syscall.result)

    # lets "close" everything that never got closed
    for fd, (filename, mmaps) in files["open"].items():
        files["closed"][filename] = mmaps

    return files["closed"]


def hook_addr(binary, addr, asm_code=None, bin_code=b""):
    main_bin = io.BytesIO(binary)
    loader = cle.Loader(main_bin, auto_load_libs=False, perform_relocations=False)
    offset = loader.main_object.addr_to_offset(addr)
    main_bin.seek(offset)
    main_bin.write(loader.main_object.arch.asm(asm_code) if asm_code else bin_code)
    main_bin.seek(0)
    return main_bin.read()



import subprocess
import tempfile
import archr
import os
import unittest


def test_hook():
    with open("/bin/false", "rb") as off:
        ofb = off.read()
    nfn = tempfile.mktemp()
    nfb = archr.utils.hook_entry(ofb, "mov rax, 0x3c; mov rdi, 0x2a; syscall")
    with open(nfn, "wb") as nff:
        nff.write(nfb)
    os.chmod(nfn, 0o755)
    assert subprocess.Popen(["/bin/false"]).wait() == 1
    assert subprocess.Popen([nfn]).wait() == 42
    os.unlink(nfn)


def test_deps():
    assert sorted(archr.utils.lib_dependencies("/bin/false")) == [
        "/lib/x86_64-linux-gnu/libc.so.6",
        "/lib64/ld-linux-x86-64.so.2",
    ]


if __name__ == "__main__":
    unittest.main()



import os
import logging
from contextlib import contextmanager

from . import ContextAnalyzer
from .strace import super_yama

l = logging.getLogger("archr.analyzers.ltrace")


class LTraceAnalyzer(ContextAnalyzer):
    """
    Returns an ltrace instance which has launched a fresh instance of the process
    """

    REQUIRED_IMPLANT = "ltrace"

    @contextmanager
    def fire_context(self, args_prefix=None, trace_args=None, **kwargs):  # pylint:disable=arguments-differ
        """
        Starts ltrace with a fresh process.
        :param trace_args: Options for ltrace
        :return: Target instance returned by run_command
        """

        fire_path = os.path.join(self.target.tmpwd, "ltrace", "fire")
        args_prefix = (args_prefix or []) + [fire_path] + (trace_args or []) + ["--"]
        with self.target.flight_context(args_prefix=args_prefix, **kwargs) as flight:
            yield flight
        flight.result = flight.process.stderr.read()  # illegal, technically


class LTraceAttachAnalyzer(ContextAnalyzer):
    """
    Returns an ltrace instance attached to a running instance of the target.
    """

    REQUIRED_IMPLANT = "ltrace"

    @contextmanager
    def fire_context(self, pid=None, trace_args=None, **kwargs):  # pylint:disable=arguments-differ
        """
        Attaches ltrace to an already existing process.
        :param pid: PID of target process
        :param trace_args: Options for ltrace
        :param kwargs: Additional arguments
        :return:
        """

        super_yama()

        fire_path = os.path.join(self.target.tmpwd, "ltrace", "fire")
        cmd_args = [fire_path] + (trace_args or []) + ["-p", "%d" % pid]
        with self.target.flight_context(args=cmd_args, **kwargs) as flight:
            yield flight
        flight.result = flight.process.stderr.read()  # illegal, technically



# pylint: disable=missing-class-docstring
import glob
import os
import platform
import shutil
import subprocess
import sys
import pathlib
from distutils.command.build import build as st_build
from distutils.util import get_platform

import pkg_resources
from setuptools import Command, setup
from setuptools.command.develop import develop as st_develop
from setuptools.errors import LibError


def _build_support_libraries():
    env = os.environ.copy()
    env["CMAKE_GENERATOR"] = "Ninja"

    curr_dir = pathlib.Path(__file__).parent.absolute()
    tcp_udp_dir = os.path.join(curr_dir, "archr", "implants", "udp_tcp_convert")
    build_dir = os.path.join(tcp_udp_dir, "build")
    if os.path.exists(build_dir):
        shutil.rmtree(build_dir)

    subprocess.run(["cmake", ".", "-Bbuild"], cwd=tcp_udp_dir, env=env, check=True)
    subprocess.run(["ninja"], cwd=build_dir, env=env, check=True)

    shutil.copy(os.path.join(build_dir, "lib", "libudp_to_tcp.so"), tcp_udp_dir)


def _clean_support_libraries():
    curr_dir = pathlib.Path(__file__).parent.absolute()
    build_dir = os.path.join(curr_dir, "archr", "implants", "udp_tcp_convert", "build")
    shutil.rmtree(build_dir, ignore_errors=True)


class build(st_build):
    def run(self, *args):
        self.execute(_build_support_libraries, (), msg="Building Archr support libraries")
        super().run(*args)


class clean_native(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        self.execute(_clean_support_libraries, (), msg="Cleaning up build directories")


class develop(st_develop):
    def run(self):
        self.run_command("build")
        super().run()


cmdclass = {
    "build": build,
    "clean_native": clean_native,
    "develop": develop,
}

try:
    from setuptools.command.editable_wheel import editable_wheel as st_editable_wheel

    class editable_wheel(st_editable_wheel):
        def run(self):
            self.run_command("build")
            super().run()

    cmdclass["editable_wheel"] = editable_wheel
except ModuleNotFoundError:
    pass

if "bdist_wheel" in sys.argv and "--plat-name" not in sys.argv:
    sys.argv.append("--plat-name")
    name = get_platform()
    if "linux" in name:
        sys.argv.append("manylinux2014_" + platform.machine())
    else:
        # https://www.python.org/dev/peps/pep-0425/
        sys.argv.append(name.replace(".", "_").replace("-", "_"))

setup(cmdclass=cmdclass)



import contextlib
import tempfile
import logging
import os

l = logging.getLogger("archr.analyzers.core_analyzer")

from . import ContextAnalyzer


class CoreResults:
    local_core_path = None
    target_core_path = None


_super_core_cmd = "echo core | docker run --rm --privileged -i ubuntu tee /proc/sys/kernel/core_pattern"


class CoreAnalyzer(ContextAnalyzer):
    """
    Runs the target and retrieves a core file. Assumes a /proc/sys/kernel/core_pattern is "core".
    """

    def __init__(self, *args, **kwargs):
        with open("/proc/sys/kernel/core_pattern", "rb") as c:
            if c.read().strip() != b"core":
                l.warning("/proc/sys/kernel/core_pattern needs to be 'core'. I am setting this system-wide.")
                os.system(_super_core_cmd)
        super().__init__(*args, **kwargs)
        if type(self.target) is not targets.DockerImageTarget:
            l.warning("When using a LocalTarget, this Analyzer will chmod 777 your CWD!!! Be careful.")

    @contextlib.contextmanager
    def fire_context(self, **kwargs):  # pylint:disable=arguments-differ
        if self.target.run_command(["chmod", "777", os.path.dirname(self.target.target_path)], user="root").wait() != 0:
            raise ArchrError("failed to chmod CWD. core will *not* drop")

        r = CoreResults()
        r.target_core_path = os.path.join(os.path.dirname(self.target.target_path), "core")
        r.local_core_path = tempfile.mktemp(prefix="arch_core_")
        try:
            with self.target.flight_context(result=r, **kwargs) as flight:
                yield flight
        finally:
            with open(r.local_core_path, "wb") as c:
                c.write(self.target.retrieve_contents(r.target_core_path))


from ..errors import ArchrError
from .. import targets
